mirror of
https://github.com/NVIDIA/cuda-samples.git
synced 2026-10-13 00:08:26 +08:00
CUDA 13.2 samples update (#432)
- Added Python samples for CUDA Python 1.0 release - Renamed top-level `Samples` directory to `cpp` to accommodate Python samples.
This commit is contained in:
18
cpp/5_Domain_Specific/HSOpticalFlow/.vscode/c_cpp_properties.json
vendored
Normal file
18
cpp/5_Domain_Specific/HSOpticalFlow/.vscode/c_cpp_properties.json
vendored
Normal file
@@ -0,0 +1,18 @@
|
||||
{
|
||||
"configurations": [
|
||||
{
|
||||
"name": "Linux",
|
||||
"includePath": [
|
||||
"${workspaceFolder}/**",
|
||||
"${workspaceFolder}/../../../Common"
|
||||
],
|
||||
"defines": [],
|
||||
"compilerPath": "/usr/local/cuda/bin/nvcc",
|
||||
"cStandard": "gnu17",
|
||||
"cppStandard": "gnu++14",
|
||||
"intelliSenseMode": "linux-gcc-x64",
|
||||
"configurationProvider": "ms-vscode.makefile-tools"
|
||||
}
|
||||
],
|
||||
"version": 4
|
||||
}
|
||||
7
cpp/5_Domain_Specific/HSOpticalFlow/.vscode/extensions.json
vendored
Normal file
7
cpp/5_Domain_Specific/HSOpticalFlow/.vscode/extensions.json
vendored
Normal file
@@ -0,0 +1,7 @@
|
||||
{
|
||||
"recommendations": [
|
||||
"nvidia.nsight-vscode-edition",
|
||||
"ms-vscode.cpptools",
|
||||
"ms-vscode.makefile-tools"
|
||||
]
|
||||
}
|
||||
44
cpp/5_Domain_Specific/HSOpticalFlow/CMakeLists.txt
Normal file
44
cpp/5_Domain_Specific/HSOpticalFlow/CMakeLists.txt
Normal file
@@ -0,0 +1,44 @@
|
||||
cmake_minimum_required(VERSION 3.20)
|
||||
|
||||
list(APPEND CMAKE_MODULE_PATH "${CMAKE_CURRENT_SOURCE_DIR}/../../../cmake/Modules")
|
||||
|
||||
project(HSOpticalFlow LANGUAGES C CXX CUDA)
|
||||
|
||||
find_package(CUDAToolkit REQUIRED)
|
||||
|
||||
set(CMAKE_POSITION_INDEPENDENT_CODE ON)
|
||||
|
||||
set(CMAKE_CUDA_ARCHITECTURES 75 80 86 87 89 90 100 110 120)
|
||||
set(CMAKE_CUDA_FLAGS "${CMAKE_CUDA_FLAGS} -Wno-deprecated-gpu-targets")
|
||||
if(ENABLE_CUDA_DEBUG)
|
||||
set(CMAKE_CUDA_FLAGS "${CMAKE_CUDA_FLAGS} -G") # enable cuda-gdb (may significantly affect performance on some targets)
|
||||
else()
|
||||
set(CMAKE_CUDA_FLAGS "${CMAKE_CUDA_FLAGS} -lineinfo") # add line information to all builds for debug tools (exclusive to -G option)
|
||||
endif()
|
||||
|
||||
# Include directories and libraries
|
||||
include_directories(../../../Common)
|
||||
|
||||
# Source file
|
||||
# Add target for HSOpticalFlow
|
||||
add_executable(HSOpticalFlow flowCUDA.cu flowGold.cpp main.cpp)
|
||||
|
||||
target_compile_options(HSOpticalFlow PRIVATE $<$<COMPILE_LANGUAGE:CUDA>:--extended-lambda>)
|
||||
|
||||
target_compile_features(HSOpticalFlow PRIVATE cxx_std_17 cuda_std_17)
|
||||
|
||||
set_target_properties(HSOpticalFlow PROPERTIES CUDA_SEPARABLE_COMPILATION ON)
|
||||
|
||||
target_include_directories(HSOpticalFlow PRIVATE
|
||||
${CUDAToolkit_INCLUDE_DIRS}
|
||||
)
|
||||
|
||||
add_custom_command(TARGET HSOpticalFlow POST_BUILD
|
||||
COMMAND ${CMAKE_COMMAND} -E copy_directory
|
||||
${CMAKE_CURRENT_SOURCE_DIR}/data
|
||||
${CMAKE_CURRENT_BINARY_DIR}/data
|
||||
)
|
||||
|
||||
# Include installation configuration
|
||||
include(${CMAKE_CURRENT_SOURCE_DIR}/../../../cmake/InstallSamples.cmake)
|
||||
setup_samples_install()
|
||||
BIN
cpp/5_Domain_Specific/HSOpticalFlow/FlowCPU.flo
Normal file
BIN
cpp/5_Domain_Specific/HSOpticalFlow/FlowCPU.flo
Normal file
Binary file not shown.
BIN
cpp/5_Domain_Specific/HSOpticalFlow/FlowGPU.flo
Normal file
BIN
cpp/5_Domain_Specific/HSOpticalFlow/FlowGPU.flo
Normal file
Binary file not shown.
34
cpp/5_Domain_Specific/HSOpticalFlow/README.md
Normal file
34
cpp/5_Domain_Specific/HSOpticalFlow/README.md
Normal file
@@ -0,0 +1,34 @@
|
||||
# HSOpticalFlow - Optical Flow
|
||||
|
||||
## Description
|
||||
|
||||
Variational optical flow estimation example. Uses textures for image operations. Shows how simple PDE solver can be accelerated with CUDA.
|
||||
|
||||
## Key Concepts
|
||||
|
||||
Image Processing, Data Parallel Algorithms
|
||||
|
||||
## Supported SM Architectures
|
||||
|
||||
[SM 5.0 ](https://developer.nvidia.com/cuda-gpus) [SM 5.2 ](https://developer.nvidia.com/cuda-gpus) [SM 5.3 ](https://developer.nvidia.com/cuda-gpus) [SM 6.0 ](https://developer.nvidia.com/cuda-gpus) [SM 6.1 ](https://developer.nvidia.com/cuda-gpus) [SM 7.0 ](https://developer.nvidia.com/cuda-gpus) [SM 7.2 ](https://developer.nvidia.com/cuda-gpus) [SM 7.5 ](https://developer.nvidia.com/cuda-gpus) [SM 8.0 ](https://developer.nvidia.com/cuda-gpus) [SM 8.6 ](https://developer.nvidia.com/cuda-gpus) [SM 8.7 ](https://developer.nvidia.com/cuda-gpus) [SM 8.9 ](https://developer.nvidia.com/cuda-gpus) [SM 9.0 ](https://developer.nvidia.com/cuda-gpus)
|
||||
|
||||
## Supported OSes
|
||||
|
||||
Linux, Windows
|
||||
|
||||
## Supported CPU Architecture
|
||||
|
||||
x86_64, armv7l
|
||||
|
||||
## CUDA APIs involved
|
||||
|
||||
### [CUDA Runtime API](http://docs.nvidia.com/cuda/cuda-runtime-api/index.html)
|
||||
cudaMalloc, cudaMemcpy, cudaMemset, cudaFree
|
||||
|
||||
## Prerequisites
|
||||
|
||||
Download and install the [CUDA Toolkit](https://developer.nvidia.com/cuda-downloads) for your corresponding platform.
|
||||
|
||||
## References (for more details)
|
||||
|
||||
[whitepaper](./doc/OpticalFlow.pdf)
|
||||
62
cpp/5_Domain_Specific/HSOpticalFlow/addKernel.cuh
Normal file
62
cpp/5_Domain_Specific/HSOpticalFlow/addKernel.cuh
Normal file
@@ -0,0 +1,62 @@
|
||||
/* Copyright (c) 2022, NVIDIA CORPORATION. All rights reserved.
|
||||
*
|
||||
* Redistribution and use in source and binary forms, with or without
|
||||
* modification, are permitted provided that the following conditions
|
||||
* are met:
|
||||
* * Redistributions of source code must retain the above copyright
|
||||
* notice, this list of conditions and the following disclaimer.
|
||||
* * Redistributions in binary form must reproduce the above copyright
|
||||
* notice, this list of conditions and the following disclaimer in the
|
||||
* documentation and/or other materials provided with the distribution.
|
||||
* * Neither the name of NVIDIA CORPORATION nor the names of its
|
||||
* contributors may be used to endorse or promote products derived
|
||||
* from this software without specific prior written permission.
|
||||
*
|
||||
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
|
||||
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
|
||||
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
|
||||
* PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
|
||||
* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
|
||||
* EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
|
||||
* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
|
||||
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
|
||||
* OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
|
||||
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
|
||||
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
|
||||
*/
|
||||
|
||||
#include "common.h"
|
||||
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
/// \brief add two vectors of size _count_
|
||||
///
|
||||
/// CUDA kernel
|
||||
/// \param[in] op1 term one
|
||||
/// \param[in] op2 term two
|
||||
/// \param[in] count vector size
|
||||
/// \param[out] sum result
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
__global__ void AddKernel(const float *op1, const float *op2, int count, float *sum)
|
||||
{
|
||||
const int pos = threadIdx.x + blockIdx.x * blockDim.x;
|
||||
|
||||
if (pos >= count)
|
||||
return;
|
||||
|
||||
sum[pos] = op1[pos] + op2[pos];
|
||||
}
|
||||
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
/// \brief add two vectors of size _count_
|
||||
/// \param[in] op1 term one
|
||||
/// \param[in] op2 term two
|
||||
/// \param[in] count vector size
|
||||
/// \param[out] sum result
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
static void Add(const float *op1, const float *op2, int count, float *sum)
|
||||
{
|
||||
dim3 threads(256);
|
||||
dim3 blocks(iDivUp(count, threads.x));
|
||||
|
||||
AddKernel<<<blocks, threads>>>(op1, op2, count, sum);
|
||||
}
|
||||
76
cpp/5_Domain_Specific/HSOpticalFlow/common.h
Normal file
76
cpp/5_Domain_Specific/HSOpticalFlow/common.h
Normal file
@@ -0,0 +1,76 @@
|
||||
/* Copyright (c) 2022, NVIDIA CORPORATION. All rights reserved.
|
||||
*
|
||||
* Redistribution and use in source and binary forms, with or without
|
||||
* modification, are permitted provided that the following conditions
|
||||
* are met:
|
||||
* * Redistributions of source code must retain the above copyright
|
||||
* notice, this list of conditions and the following disclaimer.
|
||||
* * Redistributions in binary form must reproduce the above copyright
|
||||
* notice, this list of conditions and the following disclaimer in the
|
||||
* documentation and/or other materials provided with the distribution.
|
||||
* * Neither the name of NVIDIA CORPORATION nor the names of its
|
||||
* contributors may be used to endorse or promote products derived
|
||||
* from this software without specific prior written permission.
|
||||
*
|
||||
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
|
||||
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
|
||||
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
|
||||
* PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
|
||||
* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
|
||||
* EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
|
||||
* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
|
||||
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
|
||||
* OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
|
||||
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
|
||||
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
|
||||
*/
|
||||
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
// Header for common includes and utility functions
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
|
||||
#ifndef COMMON_H
|
||||
#define COMMON_H
|
||||
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
// Common includes
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
|
||||
#include <helper_cuda.h>
|
||||
#include <math.h>
|
||||
#include <memory.h>
|
||||
#include <stdio.h>
|
||||
#include <stdlib.h>
|
||||
#include <time.h>
|
||||
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
// Common constants
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
const int StrideAlignment = 32;
|
||||
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
// Common functions
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
|
||||
// Align up n to the nearest multiple of m
|
||||
inline int iAlignUp(int n, int m = StrideAlignment)
|
||||
{
|
||||
int mod = n % m;
|
||||
|
||||
if (mod)
|
||||
return n + m - mod;
|
||||
else
|
||||
return n;
|
||||
}
|
||||
|
||||
// round up n/m
|
||||
inline int iDivUp(int n, int m) { return (n + m - 1) / m; }
|
||||
|
||||
// swap two values
|
||||
template <typename T> inline void Swap(T &a, T &b)
|
||||
{
|
||||
T t = a;
|
||||
a = b;
|
||||
b = t;
|
||||
}
|
||||
#endif
|
||||
1540
cpp/5_Domain_Specific/HSOpticalFlow/data/frame10.ppm
Normal file
1540
cpp/5_Domain_Specific/HSOpticalFlow/data/frame10.ppm
Normal file
File diff suppressed because one or more lines are too long
1464
cpp/5_Domain_Specific/HSOpticalFlow/data/frame11.ppm
Normal file
1464
cpp/5_Domain_Specific/HSOpticalFlow/data/frame11.ppm
Normal file
File diff suppressed because one or more lines are too long
147
cpp/5_Domain_Specific/HSOpticalFlow/derivativesKernel.cuh
Normal file
147
cpp/5_Domain_Specific/HSOpticalFlow/derivativesKernel.cuh
Normal file
@@ -0,0 +1,147 @@
|
||||
/* Copyright (c) 2022, NVIDIA CORPORATION. All rights reserved.
|
||||
*
|
||||
* Redistribution and use in source and binary forms, with or without
|
||||
* modification, are permitted provided that the following conditions
|
||||
* are met:
|
||||
* * Redistributions of source code must retain the above copyright
|
||||
* notice, this list of conditions and the following disclaimer.
|
||||
* * Redistributions in binary form must reproduce the above copyright
|
||||
* notice, this list of conditions and the following disclaimer in the
|
||||
* documentation and/or other materials provided with the distribution.
|
||||
* * Neither the name of NVIDIA CORPORATION nor the names of its
|
||||
* contributors may be used to endorse or promote products derived
|
||||
* from this software without specific prior written permission.
|
||||
*
|
||||
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
|
||||
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
|
||||
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
|
||||
* PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
|
||||
* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
|
||||
* EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
|
||||
* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
|
||||
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
|
||||
* OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
|
||||
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
|
||||
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
|
||||
*/
|
||||
|
||||
#include "common.h"
|
||||
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
/// \brief compute image derivatives
|
||||
///
|
||||
/// CUDA kernel, relies heavily on texture unit
|
||||
/// \param[in] width image width
|
||||
/// \param[in] height image height
|
||||
/// \param[in] stride image stride
|
||||
/// \param[out] Ix x derivative
|
||||
/// \param[out] Iy y derivative
|
||||
/// \param[out] Iz temporal derivative
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
__global__ void ComputeDerivativesKernel(int width,
|
||||
int height,
|
||||
int stride,
|
||||
float *Ix,
|
||||
float *Iy,
|
||||
float *Iz,
|
||||
cudaTextureObject_t texSource,
|
||||
cudaTextureObject_t texTarget)
|
||||
{
|
||||
const int ix = threadIdx.x + blockIdx.x * blockDim.x;
|
||||
const int iy = threadIdx.y + blockIdx.y * blockDim.y;
|
||||
|
||||
const int pos = ix + iy * stride;
|
||||
|
||||
if (ix >= width || iy >= height)
|
||||
return;
|
||||
|
||||
float dx = 1.0f / (float)width;
|
||||
float dy = 1.0f / (float)height;
|
||||
|
||||
float x = ((float)ix + 0.5f) * dx;
|
||||
float y = ((float)iy + 0.5f) * dy;
|
||||
|
||||
float t0, t1;
|
||||
// x derivative
|
||||
t0 = tex2D<float>(texSource, x - 2.0f * dx, y);
|
||||
t0 -= tex2D<float>(texSource, x - 1.0f * dx, y) * 8.0f;
|
||||
t0 += tex2D<float>(texSource, x + 1.0f * dx, y) * 8.0f;
|
||||
t0 -= tex2D<float>(texSource, x + 2.0f * dx, y);
|
||||
t0 /= 12.0f;
|
||||
|
||||
t1 = tex2D<float>(texTarget, x - 2.0f * dx, y);
|
||||
t1 -= tex2D<float>(texTarget, x - 1.0f * dx, y) * 8.0f;
|
||||
t1 += tex2D<float>(texTarget, x + 1.0f * dx, y) * 8.0f;
|
||||
t1 -= tex2D<float>(texTarget, x + 2.0f * dx, y);
|
||||
t1 /= 12.0f;
|
||||
|
||||
Ix[pos] = (t0 + t1) * 0.5f;
|
||||
|
||||
// t derivative
|
||||
Iz[pos] = tex2D<float>(texTarget, x, y) - tex2D<float>(texSource, x, y);
|
||||
|
||||
// y derivative
|
||||
t0 = tex2D<float>(texSource, x, y - 2.0f * dy);
|
||||
t0 -= tex2D<float>(texSource, x, y - 1.0f * dy) * 8.0f;
|
||||
t0 += tex2D<float>(texSource, x, y + 1.0f * dy) * 8.0f;
|
||||
t0 -= tex2D<float>(texSource, x, y + 2.0f * dy);
|
||||
t0 /= 12.0f;
|
||||
|
||||
t1 = tex2D<float>(texTarget, x, y - 2.0f * dy);
|
||||
t1 -= tex2D<float>(texTarget, x, y - 1.0f * dy) * 8.0f;
|
||||
t1 += tex2D<float>(texTarget, x, y + 1.0f * dy) * 8.0f;
|
||||
t1 -= tex2D<float>(texTarget, x, y + 2.0f * dy);
|
||||
t1 /= 12.0f;
|
||||
|
||||
Iy[pos] = (t0 + t1) * 0.5f;
|
||||
}
|
||||
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
/// \brief compute image derivatives
|
||||
///
|
||||
/// \param[in] I0 source image
|
||||
/// \param[in] I1 tracked image
|
||||
/// \param[in] w image width
|
||||
/// \param[in] h image height
|
||||
/// \param[in] s image stride
|
||||
/// \param[out] Ix x derivative
|
||||
/// \param[out] Iy y derivative
|
||||
/// \param[out] Iz temporal derivative
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
static void ComputeDerivatives(const float *I0, const float *I1, int w, int h, int s, float *Ix, float *Iy, float *Iz)
|
||||
{
|
||||
dim3 threads(32, 6);
|
||||
dim3 blocks(iDivUp(w, threads.x), iDivUp(h, threads.y));
|
||||
|
||||
cudaTextureObject_t texSource, texTarget;
|
||||
cudaResourceDesc texRes;
|
||||
memset(&texRes, 0, sizeof(cudaResourceDesc));
|
||||
|
||||
texRes.resType = cudaResourceTypePitch2D;
|
||||
texRes.res.pitch2D.devPtr = (void *)I0;
|
||||
texRes.res.pitch2D.desc = cudaCreateChannelDesc<float>();
|
||||
texRes.res.pitch2D.width = w;
|
||||
texRes.res.pitch2D.height = h;
|
||||
texRes.res.pitch2D.pitchInBytes = s * sizeof(float);
|
||||
|
||||
cudaTextureDesc texDescr;
|
||||
memset(&texDescr, 0, sizeof(cudaTextureDesc));
|
||||
|
||||
texDescr.normalizedCoords = true;
|
||||
texDescr.filterMode = cudaFilterModeLinear;
|
||||
texDescr.addressMode[0] = cudaAddressModeMirror;
|
||||
texDescr.addressMode[1] = cudaAddressModeMirror;
|
||||
texDescr.readMode = cudaReadModeElementType;
|
||||
|
||||
checkCudaErrors(cudaCreateTextureObject(&texSource, &texRes, &texDescr, NULL));
|
||||
memset(&texRes, 0, sizeof(cudaResourceDesc));
|
||||
texRes.resType = cudaResourceTypePitch2D;
|
||||
texRes.res.pitch2D.devPtr = (void *)I1;
|
||||
texRes.res.pitch2D.desc = cudaCreateChannelDesc<float>();
|
||||
texRes.res.pitch2D.width = w;
|
||||
texRes.res.pitch2D.height = h;
|
||||
texRes.res.pitch2D.pitchInBytes = s * sizeof(float);
|
||||
checkCudaErrors(cudaCreateTextureObject(&texTarget, &texRes, &texDescr, NULL));
|
||||
|
||||
ComputeDerivativesKernel<<<blocks, threads>>>(w, h, s, Ix, Iy, Iz, texSource, texTarget);
|
||||
}
|
||||
BIN
cpp/5_Domain_Specific/HSOpticalFlow/doc/OpticalFlow.docx
Normal file
BIN
cpp/5_Domain_Specific/HSOpticalFlow/doc/OpticalFlow.docx
Normal file
Binary file not shown.
BIN
cpp/5_Domain_Specific/HSOpticalFlow/doc/OpticalFlow.pdf
Normal file
BIN
cpp/5_Domain_Specific/HSOpticalFlow/doc/OpticalFlow.pdf
Normal file
Binary file not shown.
97
cpp/5_Domain_Specific/HSOpticalFlow/downscaleKernel.cuh
Normal file
97
cpp/5_Domain_Specific/HSOpticalFlow/downscaleKernel.cuh
Normal file
@@ -0,0 +1,97 @@
|
||||
/* Copyright (c) 2022, NVIDIA CORPORATION. All rights reserved.
|
||||
*
|
||||
* Redistribution and use in source and binary forms, with or without
|
||||
* modification, are permitted provided that the following conditions
|
||||
* are met:
|
||||
* * Redistributions of source code must retain the above copyright
|
||||
* notice, this list of conditions and the following disclaimer.
|
||||
* * Redistributions in binary form must reproduce the above copyright
|
||||
* notice, this list of conditions and the following disclaimer in the
|
||||
* documentation and/or other materials provided with the distribution.
|
||||
* * Neither the name of NVIDIA CORPORATION nor the names of its
|
||||
* contributors may be used to endorse or promote products derived
|
||||
* from this software without specific prior written permission.
|
||||
*
|
||||
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
|
||||
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
|
||||
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
|
||||
* PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
|
||||
* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
|
||||
* EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
|
||||
* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
|
||||
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
|
||||
* OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
|
||||
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
|
||||
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
|
||||
*/
|
||||
|
||||
#include "common.h"
|
||||
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
/// \brief downscale image
|
||||
///
|
||||
/// CUDA kernel, relies heavily on texture unit
|
||||
/// \param[in] width image width
|
||||
/// \param[in] height image height
|
||||
/// \param[in] stride image stride
|
||||
/// \param[out] out result
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
__global__ void DownscaleKernel(int width, int height, int stride, float *out, cudaTextureObject_t texFine)
|
||||
{
|
||||
const int ix = threadIdx.x + blockIdx.x * blockDim.x;
|
||||
const int iy = threadIdx.y + blockIdx.y * blockDim.y;
|
||||
|
||||
if (ix >= width || iy >= height) {
|
||||
return;
|
||||
}
|
||||
|
||||
float dx = 1.0f / (float)width;
|
||||
float dy = 1.0f / (float)height;
|
||||
|
||||
float x = ((float)ix + 0.5f) * dx;
|
||||
float y = ((float)iy + 0.5f) * dy;
|
||||
|
||||
out[ix + iy * stride] = 0.25f
|
||||
* (tex2D<float>(texFine, x - dx * 0.25f, y) + tex2D<float>(texFine, x + dx * 0.25f, y)
|
||||
+ tex2D<float>(texFine, x, y - dy * 0.25f) + tex2D<float>(texFine, x, y + dy * 0.25f));
|
||||
}
|
||||
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
/// \brief downscale image
|
||||
///
|
||||
/// \param[in] src image to downscale
|
||||
/// \param[in] width image width
|
||||
/// \param[in] height image height
|
||||
/// \param[in] stride image stride
|
||||
/// \param[out] out result
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
static void
|
||||
Downscale(const float *src, int width, int height, int stride, int newWidth, int newHeight, int newStride, float *out)
|
||||
{
|
||||
dim3 threads(32, 8);
|
||||
dim3 blocks(iDivUp(newWidth, threads.x), iDivUp(newHeight, threads.y));
|
||||
|
||||
cudaTextureObject_t texFine;
|
||||
cudaResourceDesc texRes;
|
||||
memset(&texRes, 0, sizeof(cudaResourceDesc));
|
||||
|
||||
texRes.resType = cudaResourceTypePitch2D;
|
||||
texRes.res.pitch2D.devPtr = (void *)src;
|
||||
texRes.res.pitch2D.desc = cudaCreateChannelDesc<float>();
|
||||
texRes.res.pitch2D.width = width;
|
||||
texRes.res.pitch2D.height = height;
|
||||
texRes.res.pitch2D.pitchInBytes = stride * sizeof(float);
|
||||
|
||||
cudaTextureDesc texDescr;
|
||||
memset(&texDescr, 0, sizeof(cudaTextureDesc));
|
||||
|
||||
texDescr.normalizedCoords = true;
|
||||
texDescr.filterMode = cudaFilterModeLinear;
|
||||
texDescr.addressMode[0] = cudaAddressModeMirror;
|
||||
texDescr.addressMode[1] = cudaAddressModeMirror;
|
||||
texDescr.readMode = cudaReadModeElementType;
|
||||
|
||||
checkCudaErrors(cudaCreateTextureObject(&texFine, &texRes, &texDescr, NULL));
|
||||
|
||||
DownscaleKernel<<<blocks, threads>>>(newWidth, newHeight, newStride, out, texFine);
|
||||
}
|
||||
253
cpp/5_Domain_Specific/HSOpticalFlow/flowCUDA.cu
Normal file
253
cpp/5_Domain_Specific/HSOpticalFlow/flowCUDA.cu
Normal file
@@ -0,0 +1,253 @@
|
||||
/* Copyright (c) 2022, NVIDIA CORPORATION. All rights reserved.
|
||||
*
|
||||
* Redistribution and use in source and binary forms, with or without
|
||||
* modification, are permitted provided that the following conditions
|
||||
* are met:
|
||||
* * Redistributions of source code must retain the above copyright
|
||||
* notice, this list of conditions and the following disclaimer.
|
||||
* * Redistributions in binary form must reproduce the above copyright
|
||||
* notice, this list of conditions and the following disclaimer in the
|
||||
* documentation and/or other materials provided with the distribution.
|
||||
* * Neither the name of NVIDIA CORPORATION nor the names of its
|
||||
* contributors may be used to endorse or promote products derived
|
||||
* from this software without specific prior written permission.
|
||||
*
|
||||
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
|
||||
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
|
||||
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
|
||||
* PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
|
||||
* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
|
||||
* EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
|
||||
* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
|
||||
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
|
||||
* OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
|
||||
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
|
||||
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
|
||||
*/
|
||||
|
||||
#include "common.h"
|
||||
|
||||
// include kernels
|
||||
#include "addKernel.cuh"
|
||||
#include "derivativesKernel.cuh"
|
||||
#include "downscaleKernel.cuh"
|
||||
#include "solverKernel.cuh"
|
||||
#include "upscaleKernel.cuh"
|
||||
#include "warpingKernel.cuh"
|
||||
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
/// \brief method logic
|
||||
///
|
||||
/// handles memory allocations, control flow
|
||||
/// \param[in] I0 source image
|
||||
/// \param[in] I1 tracked image
|
||||
/// \param[in] width images width
|
||||
/// \param[in] height images height
|
||||
/// \param[in] stride images stride
|
||||
/// \param[in] alpha degree of displacement field smoothness
|
||||
/// \param[in] nLevels number of levels in a pyramid
|
||||
/// \param[in] nWarpIters number of warping iterations per pyramid level
|
||||
/// \param[in] nSolverIters number of solver iterations (Jacobi iterations)
|
||||
/// \param[out] u horizontal displacement
|
||||
/// \param[out] v vertical displacement
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
void ComputeFlowCUDA(const float *I0,
|
||||
const float *I1,
|
||||
int width,
|
||||
int height,
|
||||
int stride,
|
||||
float alpha,
|
||||
int nLevels,
|
||||
int nWarpIters,
|
||||
int nSolverIters,
|
||||
float *u,
|
||||
float *v)
|
||||
{
|
||||
printf("Computing optical flow on GPU...\n");
|
||||
|
||||
// pI0 and pI1 will hold device pointers
|
||||
const float **pI0 = new const float *[nLevels];
|
||||
const float **pI1 = new const float *[nLevels];
|
||||
|
||||
int *pW = new int[nLevels];
|
||||
int *pH = new int[nLevels];
|
||||
int *pS = new int[nLevels];
|
||||
|
||||
// device memory pointers
|
||||
float *d_tmp;
|
||||
float *d_du0;
|
||||
float *d_dv0;
|
||||
float *d_du1;
|
||||
float *d_dv1;
|
||||
|
||||
float *d_Ix;
|
||||
float *d_Iy;
|
||||
float *d_Iz;
|
||||
|
||||
float *d_u;
|
||||
float *d_v;
|
||||
float *d_nu;
|
||||
float *d_nv;
|
||||
|
||||
const int dataSize = stride * height * sizeof(float);
|
||||
|
||||
checkCudaErrors(cudaMalloc(&d_tmp, dataSize));
|
||||
checkCudaErrors(cudaMalloc(&d_du0, dataSize));
|
||||
checkCudaErrors(cudaMalloc(&d_dv0, dataSize));
|
||||
checkCudaErrors(cudaMalloc(&d_du1, dataSize));
|
||||
checkCudaErrors(cudaMalloc(&d_dv1, dataSize));
|
||||
|
||||
checkCudaErrors(cudaMalloc(&d_Ix, dataSize));
|
||||
checkCudaErrors(cudaMalloc(&d_Iy, dataSize));
|
||||
checkCudaErrors(cudaMalloc(&d_Iz, dataSize));
|
||||
|
||||
checkCudaErrors(cudaMalloc(&d_u, dataSize));
|
||||
checkCudaErrors(cudaMalloc(&d_v, dataSize));
|
||||
checkCudaErrors(cudaMalloc(&d_nu, dataSize));
|
||||
checkCudaErrors(cudaMalloc(&d_nv, dataSize));
|
||||
|
||||
// prepare pyramid
|
||||
|
||||
int currentLevel = nLevels - 1;
|
||||
// allocate GPU memory for input images
|
||||
checkCudaErrors(cudaMalloc(pI0 + currentLevel, dataSize));
|
||||
checkCudaErrors(cudaMalloc(pI1 + currentLevel, dataSize));
|
||||
|
||||
checkCudaErrors(cudaMemcpy((void *)pI0[currentLevel], I0, dataSize, cudaMemcpyHostToDevice));
|
||||
checkCudaErrors(cudaMemcpy((void *)pI1[currentLevel], I1, dataSize, cudaMemcpyHostToDevice));
|
||||
|
||||
pW[currentLevel] = width;
|
||||
pH[currentLevel] = height;
|
||||
pS[currentLevel] = stride;
|
||||
|
||||
for (; currentLevel > 0; --currentLevel) {
|
||||
int nw = pW[currentLevel] / 2;
|
||||
int nh = pH[currentLevel] / 2;
|
||||
int ns = iAlignUp(nw);
|
||||
|
||||
checkCudaErrors(cudaMalloc(pI0 + currentLevel - 1, ns * nh * sizeof(float)));
|
||||
checkCudaErrors(cudaMalloc(pI1 + currentLevel - 1, ns * nh * sizeof(float)));
|
||||
|
||||
Downscale(pI0[currentLevel],
|
||||
pW[currentLevel],
|
||||
pH[currentLevel],
|
||||
pS[currentLevel],
|
||||
nw,
|
||||
nh,
|
||||
ns,
|
||||
(float *)pI0[currentLevel - 1]);
|
||||
|
||||
Downscale(pI1[currentLevel],
|
||||
pW[currentLevel],
|
||||
pH[currentLevel],
|
||||
pS[currentLevel],
|
||||
nw,
|
||||
nh,
|
||||
ns,
|
||||
(float *)pI1[currentLevel - 1]);
|
||||
|
||||
pW[currentLevel - 1] = nw;
|
||||
pH[currentLevel - 1] = nh;
|
||||
pS[currentLevel - 1] = ns;
|
||||
}
|
||||
|
||||
checkCudaErrors(cudaMemset(d_u, 0, stride * height * sizeof(float)));
|
||||
checkCudaErrors(cudaMemset(d_v, 0, stride * height * sizeof(float)));
|
||||
|
||||
// compute flow
|
||||
for (; currentLevel < nLevels; ++currentLevel) {
|
||||
for (int warpIter = 0; warpIter < nWarpIters; ++warpIter) {
|
||||
checkCudaErrors(cudaMemset(d_du0, 0, dataSize));
|
||||
checkCudaErrors(cudaMemset(d_dv0, 0, dataSize));
|
||||
|
||||
checkCudaErrors(cudaMemset(d_du1, 0, dataSize));
|
||||
checkCudaErrors(cudaMemset(d_dv1, 0, dataSize));
|
||||
|
||||
// on current level we compute optical flow
|
||||
// between frame 0 and warped frame 1
|
||||
WarpImage(pI1[currentLevel], pW[currentLevel], pH[currentLevel], pS[currentLevel], d_u, d_v, d_tmp);
|
||||
|
||||
ComputeDerivatives(
|
||||
pI0[currentLevel], d_tmp, pW[currentLevel], pH[currentLevel], pS[currentLevel], d_Ix, d_Iy, d_Iz);
|
||||
|
||||
for (int iter = 0; iter < nSolverIters; ++iter) {
|
||||
SolveForUpdate(d_du0,
|
||||
d_dv0,
|
||||
d_Ix,
|
||||
d_Iy,
|
||||
d_Iz,
|
||||
pW[currentLevel],
|
||||
pH[currentLevel],
|
||||
pS[currentLevel],
|
||||
alpha,
|
||||
d_du1,
|
||||
d_dv1);
|
||||
|
||||
Swap(d_du0, d_du1);
|
||||
Swap(d_dv0, d_dv1);
|
||||
}
|
||||
|
||||
// update u, v
|
||||
Add(d_u, d_du0, pH[currentLevel] * pS[currentLevel], d_u);
|
||||
Add(d_v, d_dv0, pH[currentLevel] * pS[currentLevel], d_v);
|
||||
}
|
||||
|
||||
if (currentLevel != nLevels - 1) {
|
||||
// prolongate solution
|
||||
float scaleX = (float)pW[currentLevel + 1] / (float)pW[currentLevel];
|
||||
|
||||
Upscale(d_u,
|
||||
pW[currentLevel],
|
||||
pH[currentLevel],
|
||||
pS[currentLevel],
|
||||
pW[currentLevel + 1],
|
||||
pH[currentLevel + 1],
|
||||
pS[currentLevel + 1],
|
||||
scaleX,
|
||||
d_nu);
|
||||
|
||||
float scaleY = (float)pH[currentLevel + 1] / (float)pH[currentLevel];
|
||||
|
||||
Upscale(d_v,
|
||||
pW[currentLevel],
|
||||
pH[currentLevel],
|
||||
pS[currentLevel],
|
||||
pW[currentLevel + 1],
|
||||
pH[currentLevel + 1],
|
||||
pS[currentLevel + 1],
|
||||
scaleY,
|
||||
d_nv);
|
||||
|
||||
Swap(d_u, d_nu);
|
||||
Swap(d_v, d_nv);
|
||||
}
|
||||
}
|
||||
|
||||
checkCudaErrors(cudaMemcpy(u, d_u, dataSize, cudaMemcpyDeviceToHost));
|
||||
checkCudaErrors(cudaMemcpy(v, d_v, dataSize, cudaMemcpyDeviceToHost));
|
||||
|
||||
// cleanup
|
||||
for (int i = 0; i < nLevels; ++i) {
|
||||
checkCudaErrors(cudaFree((void *)pI0[i]));
|
||||
checkCudaErrors(cudaFree((void *)pI1[i]));
|
||||
}
|
||||
|
||||
delete[] pI0;
|
||||
delete[] pI1;
|
||||
delete[] pW;
|
||||
delete[] pH;
|
||||
delete[] pS;
|
||||
|
||||
checkCudaErrors(cudaFree(d_tmp));
|
||||
checkCudaErrors(cudaFree(d_du0));
|
||||
checkCudaErrors(cudaFree(d_dv0));
|
||||
checkCudaErrors(cudaFree(d_du1));
|
||||
checkCudaErrors(cudaFree(d_dv1));
|
||||
checkCudaErrors(cudaFree(d_Ix));
|
||||
checkCudaErrors(cudaFree(d_Iy));
|
||||
checkCudaErrors(cudaFree(d_Iz));
|
||||
checkCudaErrors(cudaFree(d_nu));
|
||||
checkCudaErrors(cudaFree(d_nv));
|
||||
checkCudaErrors(cudaFree(d_u));
|
||||
checkCudaErrors(cudaFree(d_v));
|
||||
}
|
||||
42
cpp/5_Domain_Specific/HSOpticalFlow/flowCUDA.h
Normal file
42
cpp/5_Domain_Specific/HSOpticalFlow/flowCUDA.h
Normal file
@@ -0,0 +1,42 @@
|
||||
/* Copyright (c) 2022, NVIDIA CORPORATION. All rights reserved.
|
||||
*
|
||||
* Redistribution and use in source and binary forms, with or without
|
||||
* modification, are permitted provided that the following conditions
|
||||
* are met:
|
||||
* * Redistributions of source code must retain the above copyright
|
||||
* notice, this list of conditions and the following disclaimer.
|
||||
* * Redistributions in binary form must reproduce the above copyright
|
||||
* notice, this list of conditions and the following disclaimer in the
|
||||
* documentation and/or other materials provided with the distribution.
|
||||
* * Neither the name of NVIDIA CORPORATION nor the names of its
|
||||
* contributors may be used to endorse or promote products derived
|
||||
* from this software without specific prior written permission.
|
||||
*
|
||||
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
|
||||
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
|
||||
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
|
||||
* PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
|
||||
* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
|
||||
* EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
|
||||
* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
|
||||
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
|
||||
* OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
|
||||
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
|
||||
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
|
||||
*/
|
||||
|
||||
#ifndef FLOW_CUDA_H
|
||||
#define FLOW_CUDA_H
|
||||
|
||||
void ComputeFlowCUDA(const float *I0, // source frame
|
||||
const float *I1, // tracked frame
|
||||
int width, // frame width
|
||||
int height, // frame height
|
||||
int stride, // row access stride
|
||||
float alpha, // smoothness coefficient
|
||||
int nLevels, // number of levels in pyramid
|
||||
int nWarpIters, // number of warping iterations per pyramid level
|
||||
int nSolverIters, // number of solver iterations (for linear system)
|
||||
float *u, // output horizontal flow
|
||||
float *v); // output vertical flow
|
||||
#endif
|
||||
523
cpp/5_Domain_Specific/HSOpticalFlow/flowGold.cpp
Normal file
523
cpp/5_Domain_Specific/HSOpticalFlow/flowGold.cpp
Normal file
@@ -0,0 +1,523 @@
|
||||
/* Copyright (c) 2022, NVIDIA CORPORATION. All rights reserved.
|
||||
*
|
||||
* Redistribution and use in source and binary forms, with or without
|
||||
* modification, are permitted provided that the following conditions
|
||||
* are met:
|
||||
* * Redistributions of source code must retain the above copyright
|
||||
* notice, this list of conditions and the following disclaimer.
|
||||
* * Redistributions in binary form must reproduce the above copyright
|
||||
* notice, this list of conditions and the following disclaimer in the
|
||||
* documentation and/or other materials provided with the distribution.
|
||||
* * Neither the name of NVIDIA CORPORATION nor the names of its
|
||||
* contributors may be used to endorse or promote products derived
|
||||
* from this software without specific prior written permission.
|
||||
*
|
||||
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
|
||||
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
|
||||
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
|
||||
* PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
|
||||
* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
|
||||
* EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
|
||||
* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
|
||||
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
|
||||
* OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
|
||||
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
|
||||
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
|
||||
*/
|
||||
|
||||
#include "flowGold.h"
|
||||
|
||||
#include "common.h"
|
||||
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
/// \brief host texture fetch
|
||||
///
|
||||
/// read from arbitrary position within image using bilinear interpolation
|
||||
/// out of range coords are mirrored
|
||||
/// \param[in] t texture raw data
|
||||
/// \param[in] w texture width
|
||||
/// \param[in] h texture height
|
||||
/// \param[in] s texture stride
|
||||
/// \param[in] x x coord of the point to fetch value at
|
||||
/// \param[in] y y coord of the point to fetch value at
|
||||
/// \return fetched value
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
inline float Tex2D(const float *t, int w, int h, int s, float x, float y)
|
||||
{
|
||||
// integer parts in floating point format
|
||||
float intPartX, intPartY;
|
||||
|
||||
// get fractional parts of coordinates
|
||||
float dx = fabsf(modff(x, &intPartX));
|
||||
float dy = fabsf(modff(y, &intPartY));
|
||||
|
||||
// assume pixels are squares
|
||||
// one of the corners
|
||||
int ix0 = (int)intPartX;
|
||||
int iy0 = (int)intPartY;
|
||||
|
||||
// mirror out-of-range position
|
||||
if (ix0 < 0)
|
||||
ix0 = abs(ix0 + 1);
|
||||
|
||||
if (iy0 < 0)
|
||||
iy0 = abs(iy0 + 1);
|
||||
|
||||
if (ix0 >= w)
|
||||
ix0 = w * 2 - ix0 - 1;
|
||||
|
||||
if (iy0 >= h)
|
||||
iy0 = h * 2 - iy0 - 1;
|
||||
|
||||
// corner which is opposite to (ix0, iy0)
|
||||
int ix1 = ix0 + 1;
|
||||
int iy1 = iy0 + 1;
|
||||
|
||||
if (ix1 >= w)
|
||||
ix1 = w * 2 - ix1 - 1;
|
||||
|
||||
if (iy1 >= h)
|
||||
iy1 = h * 2 - iy1 - 1;
|
||||
|
||||
float res = t[ix0 + iy0 * s] * (1.0f - dx) * (1.0f - dy);
|
||||
res += t[ix1 + iy0 * s] * dx * (1.0f - dy);
|
||||
res += t[ix0 + iy1 * s] * (1.0f - dx) * dy;
|
||||
res += t[ix1 + iy1 * s] * dx * dy;
|
||||
|
||||
return res;
|
||||
}
|
||||
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
/// \brief host texture fetch
|
||||
///
|
||||
/// read specific texel value
|
||||
/// out of range coords are mirrored
|
||||
/// \param[in] t texture raw data
|
||||
/// \param[in] w texture width
|
||||
/// \param[in] h texture height
|
||||
/// \param[in] s texture stride
|
||||
/// \param[in] x x coord of the point to fetch value at
|
||||
/// \param[in] y y coord of the point to fetch value at
|
||||
/// \return fetched value
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
inline float Tex2Di(const float *src, int w, int h, int s, int x, int y)
|
||||
{
|
||||
if (x < 0)
|
||||
x = abs(x + 1);
|
||||
|
||||
if (y < 0)
|
||||
y = abs(y + 1);
|
||||
|
||||
if (x >= w)
|
||||
x = w * 2 - x - 1;
|
||||
|
||||
if (y >= h)
|
||||
y = h * 2 - y - 1;
|
||||
|
||||
return src[x + y * s];
|
||||
}
|
||||
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
/// \brief resize image
|
||||
/// \param[in] src image to downscale
|
||||
/// \param[in] width image width
|
||||
/// \param[in] height image height
|
||||
/// \param[in] stride image stride
|
||||
/// \param[in] newWidth image new width
|
||||
/// \param[in] newHeight image new height
|
||||
/// \param[in] newStride image new stride
|
||||
/// \param[out] out downscaled image data
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
static void
|
||||
Downscale(const float *src, int width, int height, int stride, int newWidth, int newHeight, int newStride, float *out)
|
||||
{
|
||||
for (int i = 0; i < newHeight; ++i) {
|
||||
for (int j = 0; j < newWidth; ++j) {
|
||||
const int srcX = j * 2;
|
||||
const int srcY = i * 2;
|
||||
// average 4 neighbouring pixels
|
||||
float sum;
|
||||
sum = Tex2Di(src, width, height, stride, srcX + 0, srcY + 0);
|
||||
sum += Tex2Di(src, width, height, stride, srcX + 0, srcY + 1);
|
||||
sum += Tex2Di(src, width, height, stride, srcX + 1, srcY + 0);
|
||||
sum += Tex2Di(src, width, height, stride, srcX + 1, srcY + 1);
|
||||
// normalize
|
||||
sum *= 0.25f;
|
||||
out[j + i * newStride] = sum;
|
||||
}
|
||||
}
|
||||
}
|
||||
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
/// \brief upscale one component of a displacement field
|
||||
/// \param[in] src field component to upscale
|
||||
/// \param[in] width field current width
|
||||
/// \param[in] height field current height
|
||||
/// \param[in] stride field current stride
|
||||
/// \param[in] newWidth field new width
|
||||
/// \param[in] newHeight field new height
|
||||
/// \param[in] newStride field new stride
|
||||
/// \param[in] scale value scale factor (multiplier)
|
||||
/// \param[out] out upscaled field component
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
static void Upscale(const float *src,
|
||||
int width,
|
||||
int height,
|
||||
int stride,
|
||||
int newWidth,
|
||||
int newHeight,
|
||||
int newStride,
|
||||
float scale,
|
||||
float *out)
|
||||
{
|
||||
for (int i = 0; i < newHeight; ++i) {
|
||||
for (int j = 0; j < newWidth; ++j) {
|
||||
// position within smaller image
|
||||
float x = ((float)j - 0.5f) * 0.5f;
|
||||
float y = ((float)i - 0.5f) * 0.5f;
|
||||
|
||||
out[j + i * newStride] = Tex2D(src, width, height, stride, x, y) * scale;
|
||||
}
|
||||
}
|
||||
}
|
||||
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
/// \brief warp image with provided vector field
|
||||
///
|
||||
/// For each output pixel there is a vector which tells which pixel
|
||||
/// from a source image should be mapped to this particular output
|
||||
/// pixel.
|
||||
/// It is assumed that images and the vector field have the same stride and
|
||||
/// resolution.
|
||||
/// \param[in] src source image
|
||||
/// \param[in] w width
|
||||
/// \param[in] h height
|
||||
/// \param[in] s stride
|
||||
/// \param[in] u horizontal displacement
|
||||
/// \param[in] v vertical displacement
|
||||
/// \param[out] out warped image
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
static void WarpImage(const float *src, int w, int h, int s, const float *u, const float *v, float *out)
|
||||
{
|
||||
for (int i = 0; i < h; ++i) {
|
||||
for (int j = 0; j < w; ++j) {
|
||||
const int pos = j + i * s;
|
||||
// warped coords
|
||||
float x = (float)j + u[pos];
|
||||
float y = (float)i + v[pos];
|
||||
|
||||
out[pos] = Tex2D(src, w, h, s, x, y);
|
||||
}
|
||||
}
|
||||
}
|
||||
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
/// \brief computes image derivatives for a pair of images
|
||||
/// \param[in] I0 source image
|
||||
/// \param[in] I1 tracked image
|
||||
/// \param[in] w images width
|
||||
/// \param[in] h images height
|
||||
/// \param[in] s images stride
|
||||
/// \param[out] Ix x derivative
|
||||
/// \param[out] Iy y derivative
|
||||
/// \param[out] Iz temporal derivative
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
static void ComputeDerivatives(const float *I0, const float *I1, int w, int h, int s, float *Ix, float *Iy, float *Iz)
|
||||
{
|
||||
for (int i = 0; i < h; ++i) {
|
||||
for (int j = 0; j < w; ++j) {
|
||||
const int pos = j + i * s;
|
||||
float t0, t1;
|
||||
// derivative filter is (1, -8, 0, 8, -1)/12
|
||||
// x derivative
|
||||
t0 = Tex2Di(I0, w, h, s, j - 2, i);
|
||||
t0 -= Tex2Di(I0, w, h, s, j - 1, i) * 8.0f;
|
||||
t0 += Tex2Di(I0, w, h, s, j + 1, i) * 8.0f;
|
||||
t0 -= Tex2Di(I0, w, h, s, j + 2, i);
|
||||
t0 /= 12.0f;
|
||||
|
||||
t1 = Tex2Di(I1, w, h, s, j - 2, i);
|
||||
t1 -= Tex2Di(I1, w, h, s, j - 1, i) * 8.0f;
|
||||
t1 += Tex2Di(I1, w, h, s, j + 1, i) * 8.0f;
|
||||
t1 -= Tex2Di(I1, w, h, s, j + 2, i);
|
||||
t1 /= 12.0f;
|
||||
|
||||
// spatial derivatives are averaged
|
||||
Ix[pos] = (t0 + t1) * 0.5f;
|
||||
|
||||
// t derivative
|
||||
Iz[pos] = I1[pos] - I0[pos];
|
||||
|
||||
// y derivative
|
||||
t0 = Tex2Di(I0, w, h, s, j, i - 2);
|
||||
t0 -= Tex2Di(I0, w, h, s, j, i - 1) * 8.0f;
|
||||
t0 += Tex2Di(I0, w, h, s, j, i + 1) * 8.0f;
|
||||
t0 -= Tex2Di(I0, w, h, s, j, i + 2);
|
||||
t0 /= 12.0f;
|
||||
|
||||
t1 = Tex2Di(I1, w, h, s, j, i - 2);
|
||||
t1 -= Tex2Di(I1, w, h, s, j, i - 1) * 8.0f;
|
||||
t1 += Tex2Di(I1, w, h, s, j, i + 1) * 8.0f;
|
||||
t1 -= Tex2Di(I1, w, h, s, j, i + 2);
|
||||
t1 /= 12.0f;
|
||||
|
||||
Iy[pos] = (t0 + t1) * 0.5f;
|
||||
}
|
||||
}
|
||||
}
|
||||
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
/// \brief one iteration of classical Horn-Schunck method
|
||||
///
|
||||
/// It is one iteration of Jacobi method for a corresponding linear system
|
||||
/// \param[in] du0 current horizontal displacement approximation
|
||||
/// \param[in] dv0 current vertical displacement approximation
|
||||
/// \param[in] Ix image x derivative
|
||||
/// \param[in] Iy image y derivative
|
||||
/// \param[in] Iz temporal derivative
|
||||
/// \param[in] w width
|
||||
/// \param[in] h height
|
||||
/// \param[in] s stride
|
||||
/// \param[in] alpha degree of smoothness
|
||||
/// \param[out] du1 new horizontal displacement approximation
|
||||
/// \param[out] dv1 new vertical displacement approximation
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
static void SolveForUpdate(const float *du0,
|
||||
const float *dv0,
|
||||
const float *Ix,
|
||||
const float *Iy,
|
||||
const float *Iz,
|
||||
int w,
|
||||
int h,
|
||||
int s,
|
||||
float alpha,
|
||||
float *du1,
|
||||
float *dv1)
|
||||
{
|
||||
for (int i = 0; i < h; ++i) {
|
||||
for (int j = 0; j < w; ++j) {
|
||||
const int pos = j + i * s;
|
||||
int left, right, up, down;
|
||||
|
||||
// handle borders
|
||||
if (j != 0)
|
||||
left = pos - 1;
|
||||
else
|
||||
left = pos;
|
||||
|
||||
if (j != w - 1)
|
||||
right = pos + 1;
|
||||
else
|
||||
right = pos;
|
||||
|
||||
if (i != 0)
|
||||
down = pos - s;
|
||||
else
|
||||
down = pos;
|
||||
|
||||
if (i != h - 1)
|
||||
up = pos + s;
|
||||
else
|
||||
up = pos;
|
||||
|
||||
float sumU = (du0[left] + du0[right] + du0[up] + du0[down]) * 0.25f;
|
||||
float sumV = (dv0[left] + dv0[right] + dv0[up] + dv0[down]) * 0.25f;
|
||||
|
||||
float frac = (Ix[pos] * sumU + Iy[pos] * sumV + Iz[pos]) / (Ix[pos] * Ix[pos] + Iy[pos] * Iy[pos] + alpha);
|
||||
|
||||
du1[pos] = sumU - Ix[pos] * frac;
|
||||
dv1[pos] = sumV - Iy[pos] * frac;
|
||||
}
|
||||
}
|
||||
}
|
||||
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
/// \brief method logic
|
||||
///
|
||||
/// handles memory allocation and control flow
|
||||
/// \param[in] I0 source image
|
||||
/// \param[in] I1 tracked image
|
||||
/// \param[in] width images width
|
||||
/// \param[in] height images height
|
||||
/// \param[in] stride images stride
|
||||
/// \param[in] alpha degree of displacement field smoothness
|
||||
/// \param[in] nLevels number of levels in a pyramid
|
||||
/// \param[in] nWarpIters number of warping iterations per pyramid level
|
||||
/// \param[in] nSolverIters number of solver iterations (Jacobi iterations)
|
||||
/// \param[out] u horizontal displacement
|
||||
/// \param[out] v vertical displacement
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
void ComputeFlowGold(const float *I0,
|
||||
const float *I1,
|
||||
int width,
|
||||
int height,
|
||||
int stride,
|
||||
float alpha,
|
||||
int nLevels,
|
||||
int nWarpIters,
|
||||
int nSolverIters,
|
||||
float *u,
|
||||
float *v)
|
||||
{
|
||||
printf("Computing optical flow on CPU...\n");
|
||||
|
||||
float *u0 = u;
|
||||
float *v0 = v;
|
||||
|
||||
const float **pI0 = new const float *[nLevels];
|
||||
const float **pI1 = new const float *[nLevels];
|
||||
|
||||
int *pW = new int[nLevels];
|
||||
int *pH = new int[nLevels];
|
||||
int *pS = new int[nLevels];
|
||||
|
||||
const int pixelCountAligned = height * stride;
|
||||
|
||||
float *tmp = new float[pixelCountAligned];
|
||||
float *du0 = new float[pixelCountAligned];
|
||||
float *dv0 = new float[pixelCountAligned];
|
||||
float *du1 = new float[pixelCountAligned];
|
||||
float *dv1 = new float[pixelCountAligned];
|
||||
float *Ix = new float[pixelCountAligned];
|
||||
float *Iy = new float[pixelCountAligned];
|
||||
float *Iz = new float[pixelCountAligned];
|
||||
float *nu = new float[pixelCountAligned];
|
||||
float *nv = new float[pixelCountAligned];
|
||||
|
||||
// prepare pyramid
|
||||
int currentLevel = nLevels - 1;
|
||||
pI0[currentLevel] = I0;
|
||||
pI1[currentLevel] = I1;
|
||||
|
||||
pW[currentLevel] = width;
|
||||
pH[currentLevel] = height;
|
||||
pS[currentLevel] = stride;
|
||||
|
||||
for (; currentLevel > 0; --currentLevel) {
|
||||
int nw = pW[currentLevel] / 2;
|
||||
int nh = pH[currentLevel] / 2;
|
||||
int ns = iAlignUp(nw);
|
||||
pI0[currentLevel - 1] = new float[ns * nh];
|
||||
pI1[currentLevel - 1] = new float[ns * nh];
|
||||
|
||||
Downscale(pI0[currentLevel],
|
||||
pW[currentLevel],
|
||||
pH[currentLevel],
|
||||
pS[currentLevel],
|
||||
nw,
|
||||
nh,
|
||||
ns,
|
||||
(float *)pI0[currentLevel - 1]);
|
||||
|
||||
Downscale(pI1[currentLevel],
|
||||
pW[currentLevel],
|
||||
pH[currentLevel],
|
||||
pS[currentLevel],
|
||||
nw,
|
||||
nh,
|
||||
ns,
|
||||
(float *)pI1[currentLevel - 1]);
|
||||
|
||||
pW[currentLevel - 1] = nw;
|
||||
pH[currentLevel - 1] = nh;
|
||||
pS[currentLevel - 1] = ns;
|
||||
}
|
||||
|
||||
// initial approximation
|
||||
memset(u, 0, stride * height * sizeof(float));
|
||||
memset(v, 0, stride * height * sizeof(float));
|
||||
|
||||
// compute flow
|
||||
for (; currentLevel < nLevels; ++currentLevel) {
|
||||
for (int warpIter = 0; warpIter < nWarpIters; ++warpIter) {
|
||||
memset(du0, 0, pixelCountAligned * sizeof(float));
|
||||
memset(dv0, 0, pixelCountAligned * sizeof(float));
|
||||
|
||||
memset(du1, 0, pixelCountAligned * sizeof(float));
|
||||
memset(dv1, 0, pixelCountAligned * sizeof(float));
|
||||
|
||||
WarpImage(pI1[currentLevel], pW[currentLevel], pH[currentLevel], pS[currentLevel], u, v, tmp);
|
||||
|
||||
// on current level we compute optical flow
|
||||
// between frame 0 and warped frame 1
|
||||
ComputeDerivatives(
|
||||
pI0[currentLevel], tmp, pW[currentLevel], pH[currentLevel], pS[currentLevel], Ix, Iy, Iz);
|
||||
|
||||
for (int iter = 0; iter < nSolverIters; ++iter) {
|
||||
SolveForUpdate(
|
||||
du0, dv0, Ix, Iy, Iz, pW[currentLevel], pH[currentLevel], pS[currentLevel], alpha, du1, dv1);
|
||||
Swap(du0, du1);
|
||||
Swap(dv0, dv1);
|
||||
}
|
||||
|
||||
// update u, v
|
||||
for (int i = 0; i < pH[currentLevel] * pS[currentLevel]; ++i) {
|
||||
u[i] += du0[i];
|
||||
v[i] += dv0[i];
|
||||
}
|
||||
} // end for (int warpIter = 0; warpIter < nWarpIters; ++warpIter)
|
||||
|
||||
if (currentLevel != nLevels - 1) {
|
||||
// prolongate solution
|
||||
float scaleX = (float)pW[currentLevel + 1] / (float)pW[currentLevel];
|
||||
|
||||
Upscale(u,
|
||||
pW[currentLevel],
|
||||
pH[currentLevel],
|
||||
pS[currentLevel],
|
||||
pW[currentLevel + 1],
|
||||
pH[currentLevel + 1],
|
||||
pS[currentLevel + 1],
|
||||
scaleX,
|
||||
nu);
|
||||
|
||||
float scaleY = (float)pH[currentLevel + 1] / (float)pH[currentLevel];
|
||||
|
||||
Upscale(v,
|
||||
pW[currentLevel],
|
||||
pH[currentLevel],
|
||||
pS[currentLevel],
|
||||
pW[currentLevel + 1],
|
||||
pH[currentLevel + 1],
|
||||
pS[currentLevel + 1],
|
||||
scaleY,
|
||||
nv);
|
||||
|
||||
Swap(u, nu);
|
||||
Swap(v, nv);
|
||||
}
|
||||
} // end for (; currentLevel < nLevels; ++currentLevel)
|
||||
|
||||
if (u != u0) {
|
||||
// solution is not in the specified array
|
||||
// copy
|
||||
memcpy(u0, u, pixelCountAligned * sizeof(float));
|
||||
memcpy(v0, v, pixelCountAligned * sizeof(float));
|
||||
Swap(u, nu);
|
||||
Swap(v, nv);
|
||||
}
|
||||
|
||||
// cleanup
|
||||
// last level is not being freed here
|
||||
// because it refers to input images
|
||||
for (int i = 0; i < nLevels - 1; ++i) {
|
||||
delete[] pI0[i];
|
||||
delete[] pI1[i];
|
||||
}
|
||||
|
||||
delete[] pI0;
|
||||
delete[] pI1;
|
||||
delete[] pW;
|
||||
delete[] pH;
|
||||
delete[] pS;
|
||||
delete[] tmp;
|
||||
delete[] du0;
|
||||
delete[] dv0;
|
||||
delete[] du1;
|
||||
delete[] dv1;
|
||||
delete[] Ix;
|
||||
delete[] Iy;
|
||||
delete[] Iz;
|
||||
delete[] nu;
|
||||
delete[] nv;
|
||||
}
|
||||
43
cpp/5_Domain_Specific/HSOpticalFlow/flowGold.h
Normal file
43
cpp/5_Domain_Specific/HSOpticalFlow/flowGold.h
Normal file
@@ -0,0 +1,43 @@
|
||||
/* Copyright (c) 2022, NVIDIA CORPORATION. All rights reserved.
|
||||
*
|
||||
* Redistribution and use in source and binary forms, with or without
|
||||
* modification, are permitted provided that the following conditions
|
||||
* are met:
|
||||
* * Redistributions of source code must retain the above copyright
|
||||
* notice, this list of conditions and the following disclaimer.
|
||||
* * Redistributions in binary form must reproduce the above copyright
|
||||
* notice, this list of conditions and the following disclaimer in the
|
||||
* documentation and/or other materials provided with the distribution.
|
||||
* * Neither the name of NVIDIA CORPORATION nor the names of its
|
||||
* contributors may be used to endorse or promote products derived
|
||||
* from this software without specific prior written permission.
|
||||
*
|
||||
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
|
||||
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
|
||||
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
|
||||
* PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
|
||||
* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
|
||||
* EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
|
||||
* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
|
||||
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
|
||||
* OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
|
||||
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
|
||||
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
|
||||
*/
|
||||
|
||||
#ifndef FLOW_GOLD_H
|
||||
#define FLOW_GOLD_H
|
||||
|
||||
void ComputeFlowGold(const float *I0, // source frame
|
||||
const float *I1, // tracked frame
|
||||
int width, // frame width
|
||||
int height, // frame height
|
||||
int stride, // row access stride
|
||||
float alpha, // smoothness coefficient
|
||||
int nLevels, // number of levels in pyramid
|
||||
int nWarpIters, // number of warping iterations per pyramid level
|
||||
int nIters, // number of solver iterations (for linear system)
|
||||
float *u, // output horizontal flow
|
||||
float *v); // output vertical flow
|
||||
|
||||
#endif
|
||||
239
cpp/5_Domain_Specific/HSOpticalFlow/main.cpp
Normal file
239
cpp/5_Domain_Specific/HSOpticalFlow/main.cpp
Normal file
@@ -0,0 +1,239 @@
|
||||
/* Copyright (c) 2022, NVIDIA CORPORATION. All rights reserved.
|
||||
*
|
||||
* Redistribution and use in source and binary forms, with or without
|
||||
* modification, are permitted provided that the following conditions
|
||||
* are met:
|
||||
* * Redistributions of source code must retain the above copyright
|
||||
* notice, this list of conditions and the following disclaimer.
|
||||
* * Redistributions in binary form must reproduce the above copyright
|
||||
* notice, this list of conditions and the following disclaimer in the
|
||||
* documentation and/or other materials provided with the distribution.
|
||||
* * Neither the name of NVIDIA CORPORATION nor the names of its
|
||||
* contributors may be used to endorse or promote products derived
|
||||
* from this software without specific prior written permission.
|
||||
*
|
||||
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
|
||||
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
|
||||
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
|
||||
* PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
|
||||
* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
|
||||
* EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
|
||||
* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
|
||||
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
|
||||
* OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
|
||||
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
|
||||
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
|
||||
*/
|
||||
|
||||
const static char *const sSDKsample = "HSOpticalFlow";
|
||||
|
||||
// CPU-GPU discrepancy threshold for self-test
|
||||
const float THRESHOLD = 0.05f;
|
||||
|
||||
#include <cuda_runtime.h>
|
||||
#include <helper_functions.h>
|
||||
|
||||
#include "common.h"
|
||||
#include "flowCUDA.h"
|
||||
#include "flowGold.h"
|
||||
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
/// \brief save optical flow in format described on vision.middlebury.edu/flow
|
||||
/// \param[in] name output file name
|
||||
/// \param[in] w optical flow field width
|
||||
/// \param[in] h optical flow field height
|
||||
/// \param[in] s optical flow field row stride
|
||||
/// \param[in] u horizontal displacement
|
||||
/// \param[in] v vertical displacement
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
void WriteFloFile(const char *name, int w, int h, int s, const float *u, const float *v)
|
||||
{
|
||||
FILE *stream;
|
||||
stream = fopen(name, "wb");
|
||||
|
||||
if (stream == 0) {
|
||||
printf("Could not save flow to \"%s\"\n", name);
|
||||
return;
|
||||
}
|
||||
|
||||
float data = 202021.25f;
|
||||
fwrite(&data, sizeof(float), 1, stream);
|
||||
fwrite(&w, sizeof(w), 1, stream);
|
||||
fwrite(&h, sizeof(h), 1, stream);
|
||||
|
||||
for (int i = 0; i < h; ++i) {
|
||||
for (int j = 0; j < w; ++j) {
|
||||
const int pos = j + i * s;
|
||||
fwrite(u + pos, sizeof(float), 1, stream);
|
||||
fwrite(v + pos, sizeof(float), 1, stream);
|
||||
}
|
||||
}
|
||||
|
||||
fclose(stream);
|
||||
}
|
||||
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
/// \brief
|
||||
/// load 4-channel unsigned byte image
|
||||
/// and convert it to single channel FP32 image
|
||||
/// \param[out] img_data pointer to raw image data
|
||||
/// \param[out] img_w image width
|
||||
/// \param[out] img_h image height
|
||||
/// \param[out] img_s image row stride
|
||||
/// \param[in] name image file name
|
||||
/// \param[in] exePath executable file path
|
||||
/// \return true if image is successfully loaded or false otherwise
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
bool LoadImageAsFP32(float *&img_data, int &img_w, int &img_h, int &img_s, const char *name, const char *exePath)
|
||||
{
|
||||
printf("Loading \"%s\" ...\n", name);
|
||||
char *name_ = sdkFindFilePath(name, exePath);
|
||||
|
||||
if (!name_) {
|
||||
printf("File not found\n");
|
||||
return false;
|
||||
}
|
||||
|
||||
unsigned char *data = 0;
|
||||
unsigned int w = 0, h = 0;
|
||||
bool result = sdkLoadPPM4ub(name_, &data, &w, &h);
|
||||
|
||||
if (result == false) {
|
||||
printf("Invalid file format\n");
|
||||
return false;
|
||||
}
|
||||
|
||||
img_w = w;
|
||||
img_h = h;
|
||||
img_s = iAlignUp(img_w);
|
||||
|
||||
img_data = new float[img_s * h];
|
||||
|
||||
// source is 4 channel image
|
||||
const int widthStep = 4 * img_w;
|
||||
|
||||
for (int i = 0; i < img_h; ++i) {
|
||||
for (int j = 0; j < img_w; ++j) {
|
||||
img_data[j + i * img_s] = ((float)data[j * 4 + i * widthStep]) / 255.0f;
|
||||
}
|
||||
}
|
||||
|
||||
return true;
|
||||
}
|
||||
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
/// \brief compare given flow field with gold (L1 norm)
|
||||
/// \param[in] width optical flow field width
|
||||
/// \param[in] height optical flow field height
|
||||
/// \param[in] stride optical flow field row stride
|
||||
/// \param[in] h_uGold horizontal displacement, gold
|
||||
/// \param[in] h_vGold vertical displacement, gold
|
||||
/// \param[in] h_u horizontal displacement
|
||||
/// \param[in] h_v vertical displacement
|
||||
/// \return true if discrepancy is lower than a given threshold
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
bool CompareWithGold(int width,
|
||||
int height,
|
||||
int stride,
|
||||
const float *h_uGold,
|
||||
const float *h_vGold,
|
||||
const float *h_u,
|
||||
const float *h_v)
|
||||
{
|
||||
float error = 0.0f;
|
||||
|
||||
for (int i = 0; i < height; ++i) {
|
||||
for (int j = 0; j < width; ++j) {
|
||||
const int pos = j + i * stride;
|
||||
error += fabsf(h_u[pos] - h_uGold[pos]) + fabsf(h_v[pos] - h_vGold[pos]);
|
||||
}
|
||||
}
|
||||
|
||||
error /= (float)(width * height);
|
||||
|
||||
printf("L1 error : %.6f\n", error);
|
||||
|
||||
return (error < THRESHOLD);
|
||||
}
|
||||
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
/// application entry point
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
int main(int argc, char **argv)
|
||||
{
|
||||
// welcome message
|
||||
printf("%s Starting...\n\n", sSDKsample);
|
||||
|
||||
// pick GPU
|
||||
findCudaDevice(argc, (const char **)argv);
|
||||
|
||||
// find images
|
||||
const char *const sourceFrameName = "frame10.ppm";
|
||||
const char *const targetFrameName = "frame11.ppm";
|
||||
|
||||
// image dimensions
|
||||
int width;
|
||||
int height;
|
||||
// row access stride
|
||||
int stride;
|
||||
|
||||
// flow is computed from source image to target image
|
||||
float *h_source; // source image, host memory
|
||||
float *h_target; // target image, host memory
|
||||
|
||||
// load image from file
|
||||
if (!LoadImageAsFP32(h_source, width, height, stride, sourceFrameName, argv[0])) {
|
||||
exit(EXIT_FAILURE);
|
||||
}
|
||||
|
||||
if (!LoadImageAsFP32(h_target, width, height, stride, targetFrameName, argv[0])) {
|
||||
exit(EXIT_FAILURE);
|
||||
}
|
||||
|
||||
// allocate host memory for CPU results
|
||||
float *h_uGold = new float[stride * height];
|
||||
float *h_vGold = new float[stride * height];
|
||||
|
||||
// allocate host memory for GPU results
|
||||
float *h_u = new float[stride * height];
|
||||
float *h_v = new float[stride * height];
|
||||
|
||||
// smoothness
|
||||
// if image brightness is not within [0,1]
|
||||
// this paramter should be scaled appropriately
|
||||
const float alpha = 0.2f;
|
||||
|
||||
// number of pyramid levels
|
||||
const int nLevels = 5;
|
||||
|
||||
// number of solver iterations on each level
|
||||
const int nSolverIters = 500;
|
||||
|
||||
// number of warping iterations
|
||||
const int nWarpIters = 3;
|
||||
|
||||
ComputeFlowGold(
|
||||
h_source, h_target, width, height, stride, alpha, nLevels, nWarpIters, nSolverIters, h_uGold, h_vGold);
|
||||
|
||||
ComputeFlowCUDA(h_source, h_target, width, height, stride, alpha, nLevels, nWarpIters, nSolverIters, h_u, h_v);
|
||||
|
||||
// compare results (L1 norm)
|
||||
bool status = CompareWithGold(width, height, stride, h_uGold, h_vGold, h_u, h_v);
|
||||
|
||||
WriteFloFile("FlowGPU.flo", width, height, stride, h_u, h_v);
|
||||
|
||||
WriteFloFile("FlowCPU.flo", width, height, stride, h_uGold, h_vGold);
|
||||
|
||||
// free resources
|
||||
delete[] h_uGold;
|
||||
delete[] h_vGold;
|
||||
|
||||
delete[] h_u;
|
||||
delete[] h_v;
|
||||
|
||||
delete[] h_source;
|
||||
delete[] h_target;
|
||||
|
||||
// report self-test status
|
||||
exit(status ? EXIT_SUCCESS : EXIT_FAILURE);
|
||||
}
|
||||
199
cpp/5_Domain_Specific/HSOpticalFlow/solverKernel.cuh
Normal file
199
cpp/5_Domain_Specific/HSOpticalFlow/solverKernel.cuh
Normal file
@@ -0,0 +1,199 @@
|
||||
/* Copyright (c) 2022, NVIDIA CORPORATION. All rights reserved.
|
||||
*
|
||||
* Redistribution and use in source and binary forms, with or without
|
||||
* modification, are permitted provided that the following conditions
|
||||
* are met:
|
||||
* * Redistributions of source code must retain the above copyright
|
||||
* notice, this list of conditions and the following disclaimer.
|
||||
* * Redistributions in binary form must reproduce the above copyright
|
||||
* notice, this list of conditions and the following disclaimer in the
|
||||
* documentation and/or other materials provided with the distribution.
|
||||
* * Neither the name of NVIDIA CORPORATION nor the names of its
|
||||
* contributors may be used to endorse or promote products derived
|
||||
* from this software without specific prior written permission.
|
||||
*
|
||||
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
|
||||
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
|
||||
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
|
||||
* PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
|
||||
* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
|
||||
* EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
|
||||
* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
|
||||
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
|
||||
* OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
|
||||
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
|
||||
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
|
||||
*/
|
||||
|
||||
#include <cooperative_groups.h>
|
||||
|
||||
#include "common.h"
|
||||
|
||||
namespace cg = cooperative_groups;
|
||||
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
/// \brief one iteration of classical Horn-Schunck method, CUDA kernel.
|
||||
///
|
||||
/// It is one iteration of Jacobi method for a corresponding linear system.
|
||||
/// Template parameters are describe CTA size
|
||||
/// \param[in] du0 current horizontal displacement approximation
|
||||
/// \param[in] dv0 current vertical displacement approximation
|
||||
/// \param[in] Ix image x derivative
|
||||
/// \param[in] Iy image y derivative
|
||||
/// \param[in] Iz temporal derivative
|
||||
/// \param[in] w width
|
||||
/// \param[in] h height
|
||||
/// \param[in] s stride
|
||||
/// \param[in] alpha degree of smoothness
|
||||
/// \param[out] du1 new horizontal displacement approximation
|
||||
/// \param[out] dv1 new vertical displacement approximation
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
template <int bx, int by>
|
||||
__global__ void JacobiIteration(const float *du0,
|
||||
const float *dv0,
|
||||
const float *Ix,
|
||||
const float *Iy,
|
||||
const float *Iz,
|
||||
int w,
|
||||
int h,
|
||||
int s,
|
||||
float alpha,
|
||||
float *du1,
|
||||
float *dv1)
|
||||
{
|
||||
// Handle to thread block group
|
||||
cg::thread_block cta = cg::this_thread_block();
|
||||
|
||||
volatile __shared__ float du[(bx + 2) * (by + 2)];
|
||||
volatile __shared__ float dv[(bx + 2) * (by + 2)];
|
||||
|
||||
const int ix = threadIdx.x + blockIdx.x * blockDim.x;
|
||||
const int iy = threadIdx.y + blockIdx.y * blockDim.y;
|
||||
|
||||
// position within global memory array
|
||||
const int pos = min(ix, w - 1) + min(iy, h - 1) * s;
|
||||
|
||||
// position within shared memory array
|
||||
const int shMemPos = threadIdx.x + 1 + (threadIdx.y + 1) * (bx + 2);
|
||||
|
||||
// Load data to shared memory.
|
||||
// load tile being processed
|
||||
du[shMemPos] = du0[pos];
|
||||
dv[shMemPos] = dv0[pos];
|
||||
|
||||
// load necessary neighbouring elements
|
||||
// We clamp out-of-range coordinates.
|
||||
// It is equivalent to mirroring
|
||||
// because we access data only one step away from borders.
|
||||
if (threadIdx.y == 0) {
|
||||
// beginning of the tile
|
||||
const int bsx = blockIdx.x * blockDim.x;
|
||||
const int bsy = blockIdx.y * blockDim.y;
|
||||
// element position within matrix
|
||||
int x, y;
|
||||
// element position within linear array
|
||||
// gm - global memory
|
||||
// sm - shared memory
|
||||
int gmPos, smPos;
|
||||
|
||||
x = min(bsx + threadIdx.x, w - 1);
|
||||
// row just below the tile
|
||||
y = max(bsy - 1, 0);
|
||||
gmPos = y * s + x;
|
||||
smPos = threadIdx.x + 1;
|
||||
du[smPos] = du0[gmPos];
|
||||
dv[smPos] = dv0[gmPos];
|
||||
|
||||
// row above the tile
|
||||
y = min(bsy + by, h - 1);
|
||||
smPos += (by + 1) * (bx + 2);
|
||||
gmPos = y * s + x;
|
||||
du[smPos] = du0[gmPos];
|
||||
dv[smPos] = dv0[gmPos];
|
||||
}
|
||||
else if (threadIdx.y == 1) {
|
||||
// beginning of the tile
|
||||
const int bsx = blockIdx.x * blockDim.x;
|
||||
const int bsy = blockIdx.y * blockDim.y;
|
||||
// element position within matrix
|
||||
int x, y;
|
||||
// element position within linear array
|
||||
// gm - global memory
|
||||
// sm - shared memory
|
||||
int gmPos, smPos;
|
||||
|
||||
y = min(bsy + threadIdx.x, h - 1);
|
||||
// column to the left
|
||||
x = max(bsx - 1, 0);
|
||||
smPos = bx + 2 + threadIdx.x * (bx + 2);
|
||||
gmPos = x + y * s;
|
||||
|
||||
// check if we are within tile
|
||||
if (threadIdx.x < by) {
|
||||
du[smPos] = du0[gmPos];
|
||||
dv[smPos] = dv0[gmPos];
|
||||
// column to the right
|
||||
x = min(bsx + bx, w - 1);
|
||||
gmPos = y * s + x;
|
||||
smPos += bx + 1;
|
||||
du[smPos] = du0[gmPos];
|
||||
dv[smPos] = dv0[gmPos];
|
||||
}
|
||||
}
|
||||
|
||||
cg::sync(cta);
|
||||
|
||||
if (ix >= w || iy >= h)
|
||||
return;
|
||||
|
||||
// now all necessary data are loaded to shared memory
|
||||
int left, right, up, down;
|
||||
left = shMemPos - 1;
|
||||
right = shMemPos + 1;
|
||||
up = shMemPos + bx + 2;
|
||||
down = shMemPos - bx - 2;
|
||||
|
||||
float sumU = (du[left] + du[right] + du[up] + du[down]) * 0.25f;
|
||||
float sumV = (dv[left] + dv[right] + dv[up] + dv[down]) * 0.25f;
|
||||
|
||||
float frac = (Ix[pos] * sumU + Iy[pos] * sumV + Iz[pos]) / (Ix[pos] * Ix[pos] + Iy[pos] * Iy[pos] + alpha);
|
||||
|
||||
du1[pos] = sumU - Ix[pos] * frac;
|
||||
dv1[pos] = sumV - Iy[pos] * frac;
|
||||
}
|
||||
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
/// \brief one iteration of classical Horn-Schunck method, CUDA kernel wrapper.
|
||||
///
|
||||
/// It is one iteration of Jacobi method for a corresponding linear system.
|
||||
/// \param[in] du0 current horizontal displacement approximation
|
||||
/// \param[in] dv0 current vertical displacement approximation
|
||||
/// \param[in] Ix image x derivative
|
||||
/// \param[in] Iy image y derivative
|
||||
/// \param[in] Iz temporal derivative
|
||||
/// \param[in] w width
|
||||
/// \param[in] h height
|
||||
/// \param[in] s stride
|
||||
/// \param[in] alpha degree of smoothness
|
||||
/// \param[out] du1 new horizontal displacement approximation
|
||||
/// \param[out] dv1 new vertical displacement approximation
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
static void SolveForUpdate(const float *du0,
|
||||
const float *dv0,
|
||||
const float *Ix,
|
||||
const float *Iy,
|
||||
const float *Iz,
|
||||
int w,
|
||||
int h,
|
||||
int s,
|
||||
float alpha,
|
||||
float *du1,
|
||||
float *dv1)
|
||||
{
|
||||
// CTA size
|
||||
dim3 threads(32, 6);
|
||||
// grid size
|
||||
dim3 blocks(iDivUp(w, threads.x), iDivUp(h, threads.y));
|
||||
|
||||
JacobiIteration<32, 6><<<blocks, threads>>>(du0, dv0, Ix, Iy, Iz, w, h, s, alpha, du1, dv1);
|
||||
}
|
||||
102
cpp/5_Domain_Specific/HSOpticalFlow/upscaleKernel.cuh
Normal file
102
cpp/5_Domain_Specific/HSOpticalFlow/upscaleKernel.cuh
Normal file
@@ -0,0 +1,102 @@
|
||||
/* Copyright (c) 2022, NVIDIA CORPORATION. All rights reserved.
|
||||
*
|
||||
* Redistribution and use in source and binary forms, with or without
|
||||
* modification, are permitted provided that the following conditions
|
||||
* are met:
|
||||
* * Redistributions of source code must retain the above copyright
|
||||
* notice, this list of conditions and the following disclaimer.
|
||||
* * Redistributions in binary form must reproduce the above copyright
|
||||
* notice, this list of conditions and the following disclaimer in the
|
||||
* documentation and/or other materials provided with the distribution.
|
||||
* * Neither the name of NVIDIA CORPORATION nor the names of its
|
||||
* contributors may be used to endorse or promote products derived
|
||||
* from this software without specific prior written permission.
|
||||
*
|
||||
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
|
||||
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
|
||||
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
|
||||
* PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
|
||||
* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
|
||||
* EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
|
||||
* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
|
||||
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
|
||||
* OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
|
||||
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
|
||||
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
|
||||
*/
|
||||
|
||||
#include "common.h"
|
||||
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
/// \brief upscale one component of a displacement field, CUDA kernel
|
||||
/// \param[in] width field width
|
||||
/// \param[in] height field height
|
||||
/// \param[in] stride field stride
|
||||
/// \param[in] scale scale factor (multiplier)
|
||||
/// \param[out] out result
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
__global__ void UpscaleKernel(int width, int height, int stride, float scale, float *out, cudaTextureObject_t texCoarse)
|
||||
{
|
||||
const int ix = threadIdx.x + blockIdx.x * blockDim.x;
|
||||
const int iy = threadIdx.y + blockIdx.y * blockDim.y;
|
||||
|
||||
if (ix >= width || iy >= height)
|
||||
return;
|
||||
|
||||
float x = ((float)ix + 0.5f) / (float)width;
|
||||
float y = ((float)iy + 0.5f) / (float)height;
|
||||
|
||||
// exploit hardware interpolation
|
||||
// and scale interpolated vector to match next pyramid level resolution
|
||||
out[ix + iy * stride] = tex2D<float>(texCoarse, x, y) * scale;
|
||||
}
|
||||
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
/// \brief upscale one component of a displacement field, kernel wrapper
|
||||
/// \param[in] src field component to upscale
|
||||
/// \param[in] width field current width
|
||||
/// \param[in] height field current height
|
||||
/// \param[in] stride field current stride
|
||||
/// \param[in] newWidth field new width
|
||||
/// \param[in] newHeight field new height
|
||||
/// \param[in] newStride field new stride
|
||||
/// \param[in] scale value scale factor (multiplier)
|
||||
/// \param[out] out upscaled field component
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
static void Upscale(const float *src,
|
||||
int width,
|
||||
int height,
|
||||
int stride,
|
||||
int newWidth,
|
||||
int newHeight,
|
||||
int newStride,
|
||||
float scale,
|
||||
float *out)
|
||||
{
|
||||
dim3 threads(32, 8);
|
||||
dim3 blocks(iDivUp(newWidth, threads.x), iDivUp(newHeight, threads.y));
|
||||
|
||||
cudaTextureObject_t texCoarse;
|
||||
cudaResourceDesc texRes;
|
||||
memset(&texRes, 0, sizeof(cudaResourceDesc));
|
||||
|
||||
texRes.resType = cudaResourceTypePitch2D;
|
||||
texRes.res.pitch2D.devPtr = (void *)src;
|
||||
texRes.res.pitch2D.desc = cudaCreateChannelDesc<float>();
|
||||
texRes.res.pitch2D.width = width;
|
||||
texRes.res.pitch2D.height = height;
|
||||
texRes.res.pitch2D.pitchInBytes = stride * sizeof(float);
|
||||
|
||||
cudaTextureDesc texDescr;
|
||||
memset(&texDescr, 0, sizeof(cudaTextureDesc));
|
||||
|
||||
texDescr.normalizedCoords = true;
|
||||
texDescr.filterMode = cudaFilterModeLinear;
|
||||
texDescr.addressMode[0] = cudaAddressModeMirror;
|
||||
texDescr.addressMode[1] = cudaAddressModeMirror;
|
||||
texDescr.readMode = cudaReadModeElementType;
|
||||
|
||||
checkCudaErrors(cudaCreateTextureObject(&texCoarse, &texRes, &texDescr, NULL));
|
||||
|
||||
UpscaleKernel<<<blocks, threads>>>(newWidth, newHeight, newStride, scale, out, texCoarse);
|
||||
}
|
||||
105
cpp/5_Domain_Specific/HSOpticalFlow/warpingKernel.cuh
Normal file
105
cpp/5_Domain_Specific/HSOpticalFlow/warpingKernel.cuh
Normal file
@@ -0,0 +1,105 @@
|
||||
/* Copyright (c) 2022, NVIDIA CORPORATION. All rights reserved.
|
||||
*
|
||||
* Redistribution and use in source and binary forms, with or without
|
||||
* modification, are permitted provided that the following conditions
|
||||
* are met:
|
||||
* * Redistributions of source code must retain the above copyright
|
||||
* notice, this list of conditions and the following disclaimer.
|
||||
* * Redistributions in binary form must reproduce the above copyright
|
||||
* notice, this list of conditions and the following disclaimer in the
|
||||
* documentation and/or other materials provided with the distribution.
|
||||
* * Neither the name of NVIDIA CORPORATION nor the names of its
|
||||
* contributors may be used to endorse or promote products derived
|
||||
* from this software without specific prior written permission.
|
||||
*
|
||||
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
|
||||
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
|
||||
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
|
||||
* PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
|
||||
* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
|
||||
* EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
|
||||
* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
|
||||
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
|
||||
* OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
|
||||
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
|
||||
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
|
||||
*/
|
||||
|
||||
#include "common.h"
|
||||
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
/// \brief warp image with a given displacement field, CUDA kernel.
|
||||
/// \param[in] width image width
|
||||
/// \param[in] height image height
|
||||
/// \param[in] stride image stride
|
||||
/// \param[in] u horizontal displacement
|
||||
/// \param[in] v vertical displacement
|
||||
/// \param[out] out result
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
__global__ void WarpingKernel(int width,
|
||||
int height,
|
||||
int stride,
|
||||
const float *u,
|
||||
const float *v,
|
||||
float *out,
|
||||
cudaTextureObject_t texToWarp)
|
||||
{
|
||||
const int ix = threadIdx.x + blockIdx.x * blockDim.x;
|
||||
const int iy = threadIdx.y + blockIdx.y * blockDim.y;
|
||||
|
||||
const int pos = ix + iy * stride;
|
||||
|
||||
if (ix >= width || iy >= height)
|
||||
return;
|
||||
|
||||
float x = ((float)ix + u[pos] + 0.5f) / (float)width;
|
||||
float y = ((float)iy + v[pos] + 0.5f) / (float)height;
|
||||
|
||||
out[pos] = tex2D<float>(texToWarp, x, y);
|
||||
}
|
||||
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
/// \brief warp image with provided vector field, CUDA kernel wrapper.
|
||||
///
|
||||
/// For each output pixel there is a vector which tells which pixel
|
||||
/// from a source image should be mapped to this particular output
|
||||
/// pixel.
|
||||
/// It is assumed that images and the vector field have the same stride and
|
||||
/// resolution.
|
||||
/// \param[in] src source image
|
||||
/// \param[in] w width
|
||||
/// \param[in] h height
|
||||
/// \param[in] s stride
|
||||
/// \param[in] u horizontal displacement
|
||||
/// \param[in] v vertical displacement
|
||||
/// \param[out] out warped image
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
static void WarpImage(const float *src, int w, int h, int s, const float *u, const float *v, float *out)
|
||||
{
|
||||
dim3 threads(32, 6);
|
||||
dim3 blocks(iDivUp(w, threads.x), iDivUp(h, threads.y));
|
||||
|
||||
cudaTextureObject_t texToWarp;
|
||||
cudaResourceDesc texRes;
|
||||
memset(&texRes, 0, sizeof(cudaResourceDesc));
|
||||
|
||||
texRes.resType = cudaResourceTypePitch2D;
|
||||
texRes.res.pitch2D.devPtr = (void *)src;
|
||||
texRes.res.pitch2D.desc = cudaCreateChannelDesc<float>();
|
||||
texRes.res.pitch2D.width = w;
|
||||
texRes.res.pitch2D.height = h;
|
||||
texRes.res.pitch2D.pitchInBytes = s * sizeof(float);
|
||||
|
||||
cudaTextureDesc texDescr;
|
||||
memset(&texDescr, 0, sizeof(cudaTextureDesc));
|
||||
|
||||
texDescr.normalizedCoords = true;
|
||||
texDescr.filterMode = cudaFilterModeLinear;
|
||||
texDescr.addressMode[0] = cudaAddressModeMirror;
|
||||
texDescr.addressMode[1] = cudaAddressModeMirror;
|
||||
texDescr.readMode = cudaReadModeElementType;
|
||||
|
||||
checkCudaErrors(cudaCreateTextureObject(&texToWarp, &texRes, &texDescr, NULL));
|
||||
|
||||
WarpingKernel<<<blocks, threads>>>(w, h, s, u, v, out, texToWarp);
|
||||
}
|
||||
Reference in New Issue
Block a user